#include "demo.i"

		.global	__vectors
		.global	endofline
		.global	de6bde5end
		.global	delay9
		.global	delay8
		.global	delay7
		.global	delayxplus16

	; Global registers:
	;   r0		Scratch
	;   r1		Scratch
	;   r2		Black (0x33)
	;   r3		0
	;   r5:r4	Sound channel 0 phase
	;   r7:r6	Sound channel 1 phase
	;   r9:r8	Sound channel 2 phase
	;   r10		Filter state, first stage
	;   r11		Filter state, second stage
	;   r13:r12	Noise shift register

	; During video lines:
	;   r14		Current copper segment number
	;   r15		Current line within copper segment (1 = bottom)
	;   r19		Current raster line (0 = top, 255 = bottom)

__vectors:
		rjmp	main			; vector 00

		; Exactly 13 words:

delay9:		nop
delay8:		nop
delay7:		ret

delayxplus16:	; 3 + 13 + x

1:		subi	r16, 3			;				1   \_ 3 * (x / 3) + 2
		brcc	1b			;				1/2 /
		neg	r16			; [1,2,3], inverse of time left	1
		sbrs	r16, 0			;				1/2 \_ 2 if set, 3 if clear
		rjmp	.			;				2   /
		sbrs	r16, 1			;				1/2 \_ 2 if set, 3 if clear
		rjmp	.			;				2   /
		sbrs	r16, 1			;				1/2 \_ 2 if set, 3 if clear
		rjmp	.			;				2   /
		ret				;				4

		; This is where vector 14 would go, but instead of an rjmp we just put the actual code here.

__vector_14:
		; Timer 0 output compare A.
		; This condition is triggered every 227 clock cycles.

		; Sometimes we want to let the main program run for much
		; longer. Then we set intskip to a non-zero value before
		; returning from the previous interrupt handler. A skipped
		; interrupt will only steal 7 + 20 = 27 cycles from the
		; main program.

		push	r16			;				2
		in	r16, SREG		;				1
		push	r16			;				2

		lds	r16, intskip		;				2
		subi	r16, 1			;				1
		brcs	i_noskip		;				1/2
i_skip:
		sts	intskip, r16		;				2
i_skip_nowr:
		pop	r16			;				2
		out	SREG, r16		;				1
		pop	r16			;				2
		reti				;				4
i_noskip:
		push	r17			;				2

		; First step is to sync with the timer.
		; Maximum interrupt latency is:
		;	4 cycles for finishing the currently executing instruction (e.g. eeprom read).
		;	4 cycles for pushing pc on stack and jumping to the interrupt vector table.
		; Then we add a 4 cycle margin for good measure.

		in	r17, TCNT0		;				1

		; r17 is in the range [17,24], indicating the current clock time.
		; We want to end up at a well-known time, so if r17 is low, we have to delay execution.

		ldi	r16, 24			;				1
		sub	r16, r17		;				1

		rcall	delayxplus16		;				16 + 24 - r17

		; Synchronized!
		; We'll call this time t43, because TCNT0 is 43 here.
		; (However, TCNT0 will be 43 at t270, t497, t724 and t951 too.)

		push	r0			;				2
		push	r1			;				2

		push	r14			;				2
		push	r15			;				2

		push	r18			;				2
		push	r19			;				2
		push	r20			;				2
		push	r21			;				2
		push	r22			;				2
		push	r23			;				2
		push	r24			;				2
		push	r25			;				2
		push	r26			;				2
		push	r27			;				2
		push	r28			;				2
		push	r29			;				2
		push	r30			;				2
		push	r31			;				2

		; Perform the state action.	; t79

		lds	r16, intnext		; Get the state number		2
		inc	r16 			;				1
		sts	intnext, r16		;				2

		cpi	r16, 6 + 1		;				1
		brlo	1f			;				1/2

		cpi	r16, 38 + 1		; t86				1
		brne	3f			;				1/2
		rjmp	i_field			; t88				2
3:
		cpi	r16, 54			; t89				1
		brlo	2f			;				1/2

		sts	intnext, r3		; t91				2
		rjmp	i_border		;				2
2:
		DE1				; t92
		rjmp	i_border		;				2
1:
		cpi	r16, 3 + 1		; t87				1
		breq	i_longlong		;				1/2

		; Action: sfx/vshort/vshort

		DE19				; t89
		mov	r17, r2			; Prepare for 2 us pulse	1
		rcall	soundroutine		; t109				77
		DE496				; t186
i_short:
		out	PORTD, r3		; t682; Pulse start		1
		DE34
		out	PORTD, r2		; t717; Pulse end (2 us)	1

		ldi	r16, 1			; Prepare skip count		1
		rjmp	i_done			; t719				2

i_longlong:
		; Action: sfx/vlong/vlong/sfx/vlong/vlong/sfx/vlong/vshort

		ldi	r19, 3			; t90				1
1:
		DE17				; t91
		mov	r17, r3			; Prepare for >2 us pulse	1
		rcall	soundroutine		; t109				77
		DE461				; t186
		out	PORTD, r2		; t647; Pulse end (30 us)	1
		DE31
		subi	r19, 1			;				1
		breq	i_short			;				1/2
		DE1
		out	PORTD, r3		; t682; Pulse start		1
		DE426

		; Take this opportunity to synchronize with Timer 1. Once
		; the system is stable (after the first field), this code
		; doesn't make a difference.

		sts	syncline, r3		;				2
		sts	TCNT1H, r3		;				2
		sts	TCNT1L, r3		;				2

		DE99
		out	PORTD, r2		; t79; Pulse end (30 us)	1
		DE9
		rjmp	1b			; t89				2

i_field:
		; Action: sfx/hsync/video x 256, then sfx/hsync

		ldi	r19, 0			; t90				1
		eor	r14, r14		;				1
field_segment:
		ldi	r31, hi8(coppertime)	; t92				1
		ldi	r30, lo8(coppertime)	;				1
		add	r30, r14		;				1
		ld	r15, Z			;				2
field_line:
		ldi	r31, hi8(fields)	; t97				1
		lds	r30, copperfield	;				2
		add	r30, r14		;				1
		lpm	r30, Z			;				3
		ldi	r31, hi8(pm(jumptable))	;				1
		rcall	setupline		; t105				137
		ijmp				; t242				2
		; (followed by 2 cycles of rjmp in the table)

de6bde5end:	DE6
		out	PORTD, r2		; t13				1
		DE7
endofline:
		DE62				; t21
		subi	r19, 0xff		; t83				1
		breq	field_done		;				1/2

		dec	r15			; t85				1
		mov	r16, r15		;				1
		tst	r16			;				1
		brne	1f			;				1/2

		inc	r14			; t89				1
		rjmp	field_segment		;				2
1:
		DE5				; t90
		rjmp	field_line		;				2

field_done:
		DE19				; t86
		rcall	setupline		; t105				137

		; It's time for a vblank pseudo-interrupt, so we'll re-enable
		; interrupts (recursively, since we're still in the timer interrupt).

		ldi	r16, 3			; t242				1
		sts	intskip, r16		;				1
		ldi	r17, 2			;				1
		out	TIFR0, r17		; Clear any pending interrupt	1

		sei

		; Now we call the vblank handler directly.
		; It may trash any registers that we've already pushed.
		; In the meantime, timer interrupts may occur and be handled.

		rcall	playroutine

		; Update the frame number

		in	r16, GPIOR1
		in	r17, GPIOR2
		ldi	r18, 1
		add	r16, r18
		adc	r17, r3
		out	GPIOR1, r16
		out	GPIOR2, r17

		ldi	r28, lo8(coppertime)
		ldi	r29, hi8(coppertime)

		ldd	r18, Y+cl_flash
		subi	r18, 0x11
		brcs	1f
		std	Y+cl_flash, r18
1:
		lds	r30, vblvector
		ldi	r31, hi8(pm(jumptable))
		icall

		; Trigger the main program if it is waiting for vblank.

		ldi	r16, 0

		sbis	GPIOR0, GFLAG_VBLANK
		rjmp	i_done

		in	r17, GPIOR0
		subi	r17, 0x81		; toggle GFLAG_DBLBUF, clear GFLAG_VBLANK
		out	GPIOR0, r17
		rjmp	i_done

i_border:
		; Action: sfx/hsync

		DE10				; t95
		rcall	setupline		; t105				137

		ldi	r16, 3			; t242; Prepare skip count	1

i_done:
		; State actions end up jumping here.
		; r16 contains the requested number of skipped interrupts.
		; Interrupts occur at t0, t227, t454, t681 and t908. Only
		; t0 interrupts should be served, the rest should be skipped.

		pop	r31			;				2
		pop	r30			;				2
		pop	r29			;				2
		pop	r28			;				2
		pop	r27			;				2
		pop	r26			;				2
		pop	r25			;				2
		pop	r24			;				2
		pop	r23			;				2
		pop	r22			;				2
		pop	r21			;				2
		pop	r20			;				2
		pop	r19			;				2
		pop	r18			;				2

		pop	r15			;				2
		pop	r14			;				2

		pop	r1			;				2
		pop	r0			;				2

		tst	r16			;				1
		breq	1f			;				1/2

		ldi	r17, 2			;				1
		out	TIFR0, r17		; Clear any pending interrupt	1
		pop	r17			;				2
		rjmp	i_skip			;				2
1:
		pop	r17			;				2
		rjmp	i_skip_nowr		;				2

soundroutine:
		; 3 + 74

		; Input:
		;	r17	What to write at t150 (used for sync generation)
		; Must preserve:
		;	r14, r15, r19, r30, r31

		lds	r0, ch+0*c_sz+c_freq	; t112				2
		add	r4, r0	      		;				1

		out	PORTD, r3		; t115; Sync start		1

		lds	r1, ch+0*c_sz+c_freq+1	;				2
		adc	r5, r1	      		;				1

		lds	r0, ch+1*c_sz+c_freq	;				2
		lds	r1, ch+1*c_sz+c_freq+1	;				2
		add	r6, r0	      		;				1
		adc	r7, r1	      		;				1

		lds	r0, ch+2*c_sz+c_freq	;				2
		lds	r1, ch+2*c_sz+c_freq+1	;				2
		add	r8, r0			;				1
		adc	r9, r1			;				1

		; Run the noise generator (four cycles!)

		lsl	r12			;				1
		rol	r13			;				1
		brvc	1f			;				\_ 2
		add	r12, r2			;				/
1:
		; Channel 1: Filtered pulse wave
		
		lds	r18, ch+1*c_sz+c_vol	; [0, 63]			2
		lds	r16, ch+1*c_sz+c_pw	;				2
		cp	r7, r16			;				1
		brcc	1f			;				\_ 2
		neg	r18			; [-63, 63]			/
1:
		; 2 pole low-pass filter

		lds	r20, ch+0*c_sz+c_pw	; cutoff			2

		sub	r18, r10		;				1
		mulsu	r18, r20		;				2
		add	r10, r1			;				1
		mov	r18, r10		; 1st stage output [-63, 63]	1

		sub	r18, r11		;				1

		out	PORTD, r17		; t150; Possible sync end	1

		mulsu	r18, r20		;				2
		add	r11, r1			;				1
		mov	r18, r11		; 2nd stage output [-63, 63]	1

		; Channel 2: Pulse wave

		lds	r0, ch+2*c_sz+c_vol	; [0, 63]			2
		lds	r16, ch+2*c_sz+c_pw	;				2
		cp	r9, r16			;				1
		brcc	1f			;				\_ 2
		neg	r0			; [-63, 63]			/
1:
		add	r18, r0			;				1
		asr	r18			; [-63, 63]			1

		; Channel 0: Triangle wave

		mov	r16, r5			;				1
		sbrc	r16, 7			;				\_ 2
		com	r16			;				/
		subi	r16, 64			; [-64, 63]			1
		lds	r20, ch+0*c_sz+c_vol	;				2
		fmuls	r16, r20		; r1 is [-32, 31]		2

		add	r18, r1			; [-95, 94]			1

		; Add noise

		lds	r0, noisevol		; [0, 31]			2
		sbrc	r13, 7			;				\_ 2
		neg	r0			;				/

		add	r18, r0			; [-126, 125]			1

		; Write to the DAC

		subi	r18, 128		; [2, 253]			1
		lsr	r18			;				1
		lsr	r18			;				1
		out	PORTC, r18		; [0, 63]			1

		ret				;				4

setupline:
		; 3 + 134

		; This routine transmits the hsync pulse and colour burst.

		mov	r17, r3			; t108; Prepare for >2 us pulse	1
		rcall	soundroutine		; t109				77

		out	PORTD, r2		; t186; Pulse ends here (4 us)	1

		lds	r16, syncline		;				2
		ldi	r17, 0x22		;				1
		ldi	r18, 0x22		;				1
		sbrc	r16, 1			;				1/2 \_ 2
		ldi	r17, 0x44		;				1   /
		sbrs	r16, 1			;				1/2 \_ 2
		ldi	r18, 0x44		;				1   /

		out	PORTD, r17		; t195
		out	PORTD, r2
		out	PORTD, r18
		out	PORTD, r2

		out	PORTD, r17
		out	PORTD, r2
		out	PORTD, r18
		out	PORTD, r2

		out	PORTD, r17
		out	PORTD, r2
		out	PORTD, r18
		out	PORTD, r2

		out	PORTD, r17
		out	PORTD, r2
		out	PORTD, r18
		out	PORTD, r2

		out	PORTD, r17
		out	PORTD, r2
		out	PORTD, r18
		out	PORTD, r2

		out	PORTD, r17
		out	PORTD, r2
		out	PORTD, r18
		out	PORTD, r2

		out	PORTD, r17
		out	PORTD, r2
		out	PORTD, r18
		out	PORTD, r2

		out	PORTD, r17
		out	PORTD, r2
		out	PORTD, r18
		out	PORTD, r2

		out	PORTD, r17
		out	PORTD, r2
		out	PORTD, r18
		out	PORTD, r2

		out	PORTD, r17
		out	PORTD, r2
		out	PORTD, r18
		out	PORTD, r2		; t234

		inc	r16			;				1
		sts	syncline, r16		;				2

		ret

